with Text_IO; use Text_IO;
with Lower_Layer_UDP;
with Ada.Strings.Unbounded;
with Ada.Text_IO;
with Ada.Streams;
with Ada.Streams.Stream_IO;
with Ada.Exceptions;
procedure Cliente is

   package LLU renames Lower_Layer_UDP;
   package ASU renames Ada.Strings.Unbounded;
   package S_IO renames Ada.Streams.Stream_IO;

   Server_EP: LLU.End_Point_Type;             -- Direccin del servidor
   Client_EP: LLU.End_Point_Type;             -- Direccion del cliente
   Opcion: Character;                         -- Para guardar si la peticin es de lectura o escritura
   Fichero: String (1..255);                  -- Nombre del fichero a leer o escribir
   Buffer:    aliased LLU.Buffer_Type(5120);
   Tipo: Natural;
   Aux: Natural;                              -- Longitud del string fichero
   Tamano_Trozo: Integer:= 512;
   Plazo_Espera: Float:= 0.2;
   Tamano_Trozo_Aux: String (1..10);
   Long_Tamano_Aux: Natural;
   Plazo_Aux: String (1..10);
   Long_Plazo_Aux: Natural;
   Salir: Boolean;
   Expired: Boolean;


   -- *************************************************************************************************
   -- **************************** PROCEDMIENTO DE PETICIN DE LECTURA O ESCRITURA ********************
   -- *************************************************************************************************

   procedure Peticion_EL is
   begin
      loop
         Opcion:=' ';
         Put ("Qu desea, escribir (e) o leer(l)? ");
         Ada.Text_IO.Get (Opcion);
         exit when (Opcion='E') or (Opcion='L') or (Opcion='e') or (Opcion='l');
      end loop;
   end Peticion_EL;



   -- ************************************************************************************************
   -- ************************* PROCEDIMIENTO PARA PEDIR EL NOMBRE DEL FICHERO ***********************
   -- ************************************************************************************************

   procedure Peticion_Fichero is
   begin
      Ada.Text_IO.Skip_Line;
      Ada.Text_IO.Put ("Qu fichero quieres leer/escribir? ");
      Ada.Text_IO.Get_Line (Fichero,aux);
   end;


   function A_Entero (U: ASU.Unbounded_String) return Integer is
      Long: Integer;
      Total: Integer;
   begin
      Long:= ASU.To_String(U)'Length;
      declare S: String := ASU.To_String(U);
      begin
         Total:= 0;
         for I in 1..Long loop
            Total:= Total* 10 +(Character'Pos(S(I))-Character'Pos('0'));
         end loop;
      end;
      return Total;
   end A_Entero;


   function A_Float (U: ASU.Unbounded_String) return Float is
      Long: Integer;
      Total:Float;
   begin
      Long:= ASU.To_String(U)'Length;
      declare
         S: String:= ASU.To_STring(U);
      begin
         Total:=0.0;
         for I in 1..Long loop
            Total:= Total*10.0 + Float ((Character'Pos(S(I)) - Character'Pos('0')));
         end loop;
      end;
      return Total;
   end A_Float;


   -- ************************************************************************************************
   -- ************************* PROCEDIMIENTO PARA PEDIR LAS POSIBLES OPCIONES ***********************
   -- ************************************************************************************************

   procedure Peticion_Opciones is
   begin
      loop
         Ada.Text_IO.Put ("Qu tamao de trozo de paquetes quieres (entre 8 y 4096)? ");
         Ada.Text_IO.Get_Line (Tamano_Trozo_Aux, Long_Tamano_Aux);
         exit when (A_Entero (ASU.To_Unbounded_String (Tamano_Trozo_Aux(1..Long_Tamano_Aux)))>7) and (A_Entero (ASU.To_Unbounded_String(Tamano_Trozo_Aux(1..Long_Tamano_Aux)))<4097);
      end loop;

      loop
         Ada.Text_IO.Put ("Qu plazo de espera quieres (en milisegundos)? ");
         Ada.Text_IO.Get_Line (Plazo_Aux, Long_Plazo_Aux);
         exit when (A_Entero (ASU.To_Unbounded_String(Plazo_Aux(1..Long_Plazo_Aux)))>99) and (A_Entero (ASU.To_Unbounded_String(Plazo_Aux(1..Long_Plazo_Aux)))<255001);
      end loop;
   end Peticion_Opciones;

   -- ************************************************************************************************
   -- ****************************** PROCEDIMIENTO DE ESCRITURA **************************************
   -- ************************************************************************************************

   procedure Escritura is
       Numero_Trozo: Natural;  -- Nmero de asentimiento de trozo que se lee del buffer
       Num_Trozo_Aux: Natural; -- Nmero del ltimo trozo asentido
       -- Trozo: Ada.Streams.Stream_Element_Array(1..Tamano_Trozo);
       Ultimo: Ada.Streams.Stream_Element_Offset; -- Indica el tamao del trozo que se lee del buffer
       Direccion_Fichero: S_IO.File_Type;  -- Variable que hace referencia al fichero donde se lee
       Aux1, Aux2, Aux_Tamano, Aux_Plazo: ASU.Unbounded_String;

   begin

      -- EL BUCLE SE REPITE HASTA RECIBIR RESPUESTA DEL SERVIDOR

      loop
         -- METO EN EL BUFFER LA PETICIN DE ESCRITURA Y LO MANDO

         LLU.Reset(Buffer);
         LLU.End_Point_Type'Output(Buffer'Access, Client_EP);
         Natural'Output (Buffer'Access,2);  -- Peticin de escritura
         ASU.Unbounded_String'Output (Buffer'Access, ASU.To_Unbounded_String (Fichero(1..Aux)));
         Natural'Output (Buffer'Access, 2);
         ASU.Unbounded_String'Output (Buffer'Access, ASU.To_Unbounded_String("Tamano"));
         ASU.Unbounded_String'Output (Buffer'Access, ASU.To_Unbounded_String(Tamano_Trozo_Aux(1..Long_Tamano_Aux)));
         ASU.Unbounded_String'Output (Buffer'Access, ASU.To_Unbounded_String("Tiempo"));
         ASU.Unbounded_String'Output (Buffer'Access, ASU.To_Unbounded_String(Plazo_Aux(1..Long_Plazo_Aux)));

         LLU.Send (Server_EP, Buffer'Access);
         Put_Line ("Mando peticin de escritura");


         -- ESPERO LA RESPUESTA DEL SERVIDOR

         LLU.Reset (Buffer);
         LLU.Receive (Client_EP, Buffer'Access, Duration (Plazo_Espera), Expired);

         -- SI RECIBO RESPUESTA ME SALGO DEL BUCLE. SI NO, SE REPITE LA ITERACIN DEL BUCLE
         -- Y VUELVE A MANDAR LA PETICIN

         if not Expired then exit; end if;
      end loop;


      --  EXTRAIGO LOS DATOS DEL BUFFER, QUE DEBEN SER EL ASENTIMIENTO DE LA ESCRITURA Y DE LAS OPCIONES

      Tipo:= Natural'Input (Buffer'Access); -- Recibo 6 del OACK
      Numero_Trozo:= Natural'Input (Buffer'Access); -- Recibo n de opciones
      Aux1:=  ASU.Unbounded_String'Input (Buffer'Access);
      Aux_Tamano:= ASU.Unbounded_String'Input (Buffer'Access);
      Aux2:= ASU.Unbounded_STring'Input (Buffer'Access);
      Aux_Plazo:= ASU.Unbounded_String'Input (Buffer'Access);
      Put_Line ("Recibo asentimiento de escritura con opciones");
      Tamano_Trozo:= A_Entero (Aux_Tamano);
      Plazo_Espera:= (A_Float (Aux_Plazo))/1000.0;


      -- ABRO EL FICHERO

      S_IO.Open (Direccion_Fichero, S_IO.In_File, Fichero(1..Aux));
      Put_Line ("Abro el fichero");

      Numero_Trozo:= 0;
      Num_Trozo_Aux:= 0;
      Salir:=False;

      declare
         Trozo: Ada.Streams.Stream_Element_Array (1..Ada.Streams.Stream_Element_Offset (Tamano_Trozo));
      begin
         -- SI EL FICHERO ES VACIO
         if S_IO.End_Of_File (Direccion_Fichero) then
            loop
               Numero_Trozo:= 1;
               LLU.Reset (Buffer);
               LLU.End_Point_Type'Output (Buffer'Access, Client_EP);
               Natural'Output (Buffer'Access, 3);
               Natural'Output (Buffer'Access, Numero_Trozo);
               Ada.Streams.Stream_Element_Array'Output (Buffer'Access, Trozo(1..0));
               LLU.Send (Server_EP, Buffer'Access);
               Put_Line ("Mando la parte del fichero" & Integer'Image (Numero_Trozo));

               LLU.Reset (Buffer);
               LLU.Receive (Client_EP, Buffer'Access, Duration (Plazo_Espera), Expired);

               -- SI HE RECIBIDO ALGO ME SALGO DEL BUCLE. SI NO, SE EJECUTA OTRA ITERACIN DEL BUCLE Y VUELVO A
               -- MANDAR EL LTIMO TROZO DE FICHERO
               if not Expired then
                  -- EXTRAIGO DEL BUFFER LOS DATOS DEL ASENTIMIENTO DEL SERVIDOR
                  Tipo:= Natural'Input (Buffer'Access);
                  Num_Trozo_Aux:= Natural'Input (Buffer'Access);
                  Put_Line ("Recibo asentimiento del trozo: " & Integer'Image (Num_Trozo_Aux));
                  if Numero_Trozo /= Num_Trozo_Aux then
                     Expired:= False;
                  end if;
               end if;

               if  not Expired then exit; end if;
            end loop;
            Salir:=True;

         end if;

         -- SE REPITE EL BUCLE HASTA QUE SE MANDEN TODOS LOS TROZOS DEL FICHERO (EXCEPTO SI EL LTIMO ES VACO)
         -- Y SEAN ASENTIDOS TODOS ESOS TROZOS

         while (not S_IO.End_Of_File (Direccion_Fichero)) or (not Salir) loop

            -- SI EL LTIMO TROZO ASENTIDO, ES EL LTIMO QUE HE LEIDO, LEO EL SIGUIENTE TROZO DE FICHERO
            -- SI NO, EN EL SIGUIENTE BUCLE, SE VUELVE A MANDAR EL LTIMO TROZO MANDADO
            if Numero_Trozo=Num_Trozo_Aux then
               Numero_Trozo:= Numero_Trozo+1;
               Num_Trozo_Aux:= Numero_Trozo-1;
               S_IO.Read (Direccion_Fichero, Trozo, Ultimo);
            end if;

            -- REPITO EL BUCLE HASTA QUE RECIBA UN ASENTIMIENTO POR PARTE DEL SERVIDOR,
            loop

               -- METO EN EL BUFFER EL TROZO DE FICHERO ACTUAL

               LLU.Reset (Buffer);
               LLU.End_Point_Type'Output(Buffer'Access, Client_EP);
               Natural'Output (Buffer'Access, 3);
               Natural'Output (Buffer'Access, Numero_Trozo);
               Ada.Streams.Stream_Element_Array'Output (Buffer'Access, Trozo(1..Ultimo));
               LLU.Send (Server_EP, Buffer'Access);
               Put_Line ("Mando la parte del fichero" & Integer'Image (Numero_Trozo));



               -- ESPERO EL ASENTIMIENTO DEL SERVIDOR

               LLU.Reset (Buffer);
               LLU.Receive (Client_EP, Buffer'Access,Duration(Plazo_Espera), Expired);

               -- SI HE RECIBIDO ALGO ME SALGO DEL BUCLE. SI NO, SE EJECUTA OTRA ITERACIN DEL BUCLE Y VUELVO A
               -- MANDAR EL LTIMO TROZO DE FICHERO

               if  not Expired then exit; end if;

            end loop;

            -- EXTRAIGO DEL BUFFER LOS DATOS DEL ASENTIMIENTO DEL SERVIDOR

            Tipo:= Natural'Input (Buffer'Access);
            Num_Trozo_Aux:= Natural'Input (Buffer'Access);
            Put_Line ("Recibo asentimiento del trozo: " & Integer'Image (Num_Trozo_Aux));

            -- SI ES FINAL DE FICHERO, Y EL TROZO ASENTIDO ES EL LTIMO TROZO MANDADO, ME SALGO DEL BUCLE

            if S_IO.End_Of_File (Direccion_Fichero) and (Num_Trozo_Aux= Numero_Trozo) then Salir:= True; end if;
         end loop;



         --COMPRUEBO SI EL LTIMO PAQUETE MANDADO ES DE Tamano_Trozo, EN ESE CASO, MANDO UN TROZO VACO, PARA INDICAR
      -- AL SERVIDOR QUE HA TERMINADO EL FICHERO

         Numero_Trozo:= Numero_Trozo+1;
         if Integer(Ultimo)=Tamano_Trozo then
            loop

               -- METO EN EL BUFFER UN TROZO DE FICHERO VACO, DE LONGITUD 0, Y LO ENVO

               LLU.Reset (Buffer);
               LLU.End_Point_Type'Output(Buffer'Access, Client_EP);
               Natural'Output (Buffer'Access, 3);
               Natural'Output (Buffer'Access, Numero_Trozo);
               Ada.Streams.Stream_Element_Array'Output (Buffer'Access, Trozo (1..0));
               LLU.Send (Server_EP, Buffer'Access);
               Put_Line ("Mando ltima parte del fichero vaca " & Integer'Image (Numero_Trozo));

               -- ESPERO EL ASENTIMIENTO DEL SERVIDOR

               LLU.Reset (Buffer);
               LLU.Receive (Client_EP, Buffer'Access,Duration (Plazo_Espera), Expired);

               -- SI RECIBO EL ASENTIMIENTO DE ESTE LTIMO TROZO ME SALGO DEL BUCLE

               if not Expired then
                  Tipo:= Natural'Input (Buffer'Access);
                  Num_Trozo_Aux:= Natural'Input (Buffer'Access);
                  if Num_Trozo_Aux = Numero_Trozo then
                     Put_Line ("Recibo asentimiento del trozo" & Integer'Image (Numero_Trozo));
                     exit;
                  end if;
               end if;
            end loop;
         end if;
      end;  -- declare

      S_IO.Close (Direccion_Fichero);


   end;


   -- **************************************************************************************************
   -- ********************************** PROCEDIMIENTO DE LECTURA **************************************
   -- **************************************************************************************************

   procedure Lectura is

      Dst_File: S_IO.File_Type;  -- Variable que se refiere al fichero donde se va a escribir
      Num_Trozo_Ant: Natural;
      Num_Trozo: Natural;
   begin
      loop
         -- METO EN EL BUFFER LA PETICIN DE LECTURA Y SE LA ENVO AL SERVIDOR

         LLU.Reset(Buffer);
         LLU.End_Point_Type'Output(Buffer'Access, Client_EP);
         Natural'Output (Buffer'Access,1);  -- Peticin de lectura
         ASU.Unbounded_String'Output (Buffer'Access, ASU.To_Unbounded_String (Fichero(1..Aux)));
         LLU.Send (Server_EP, Buffer'Access);
         Put_Line ("Mando peticin de lectura");

         -- ESPERAMOS RESPUESTA DEL SERVIDOR

         LLU.Reset (Buffer);
         LLU.Receive (Client_EP, Buffer'Access,Duration(Plazo_Espera), Expired);

         -- SI RECIBO ALGO ME SALGO DEL BUCLE. SI NO, SE REPITE LA ITERACIN DEL BUCLE Y SE VUELVO A
         -- ENVIAR LA PETICIN DE LECTURA

         if not Expired then exit; end if;
      end loop;


      -- CREO EL FICHERO PARA ESCRIBIR LO QUE ME MANDE EL SERVIDOR

      S_IO.Create(Dst_File, S_IO.Out_File, Fichero(1..Aux)&"2");



      Num_Trozo_Ant:=0;
      Tipo:= Natural'Input (Buffer'Access);

      -- EL BUCLE SE REPITE HASTA QUE RECIBE TODOS LOS TROZOS DEL FICHERO, Y SE SALE CUANDO RECIBE EL
      -- LTIMO QUE DEBE SER MENOR DE Tamano_Trozo bytes

      loop

         -- SACO LOS DATOS DEL BUFER QUE ME HA ENVIADO EL SERVIDOR

         Num_Trozo:= Natural'Input (Buffer'Access);
         Put_Line ("Recibo trozo :" & Integer'Image (Num_Trozo));
         declare
            Trozo: Ada.Streams.Stream_Element_Array:= Ada.Streams.Stream_Element_Array'Input (Buffer'Access);

         begin

            -- SI EL TROZO RECIBIDO ES EL SIGUIENTE DEL FICHERO, LO ESCRIBO

            if Num_Trozo_Ant=Num_Trozo-1 then
               S_IO.Write (Dst_File, Trozo);
               Num_Trozo_Ant:= Num_Trozo;
            end if;

            -- ESTE BUCLE SE REPITE HASTA QUE RECIBO ALGN PAQUETE DEL SERVIDOR, O CUANDO EL LTIMO TROZO
            -- ESCRITO EN EL FICHERO ERA MENOR DE Tamano_Trozo bytes

            loop
               -- MANDO ASENTIMIENTO DEL LTIMO TROZO RECIBIDO

               LLU.Reset (Buffer);
               LLU.End_Point_Type'Output (Buffer'Access, Client_EP);
               Natural'Output (Buffer'Access, 4); -- Paquete de asentimiento
               Natural'Output (Buffer'Access, Num_Trozo);
               LLU.Send (Server_EP, Buffer'Access);
               Put_Line ("Mando asentimiento del trozo " & Integer'Image (Num_Trozo));

               -- ME QUEDO ESPERANDO PARA RECIBIR OTRO TROZO DE FICHERO

               LLU.Reset (Buffer);
               LLU.Receive (Client_EP, Buffer'Access,Duration (Plazo_Espera), Expired);

               -- SI NO RECIBO NADA Y EL LTIMO TROZO ESCRITO ERA MENOR DE Tamano_Trozo bytes ME SALGO DEL BUCLE
               -- SI RECIBO UN PAQUETE TAMBIN ME SALGO DEL BUCLE

               if Expired then
                  if Trozo'Length <Tamano_Trozo then
                     Salir:= True;
                     exit;
                  end if;
               else
                  exit;

               end if;
            end loop;
         end; -- FIN DEL DECLARE

         -- SI EL TROZO ESCRITO ERA EL LTIMO, ME SALGO DEL BUCLE
         -- SI NO ERA EL LTIMO, VUELVE LA ITERACIN DEL BUCLE Y COMPRUEBA LOS DATOS QUE LE HA
         -- ENVIADO EL SERVIDOR

         if Salir then exit;
         else
            Tipo:= Natural'Input (Buffer'Access);
         end if;
      End loop;

      -- EN ESTE BUCLE SE ESPERA POR SI SE HA PERDIDO EL LTIMO ASENTIMIENTO, SI RECIBO EL LTIMO TROZO
      -- DEL SERVIDOR DE NUEVO, VUELVO A MANDAR EL ASENTIMIENTO, Y VUELVO A ESPERAR HASTA NO RECIBIR NADA
      -- DURANTE 5 SEGUNDOS. DESPUS CIERRO EL FICHERO Y DOY POR TERMINADA LA CONEXIN

      loop
         LLU.Reset (Buffer);
         LLU.Receive (Client_EP, Buffer'Access,Duration (Plazo_Espera), Expired);
         exit when Expired;
         LLU.Reset (Buffer);
         LLU.End_Point_Type'Output (Buffer'Access, Client_EP);
         Natural'Output (Buffer'Access, 4); -- Paquete de asentimiento
         Natural'Output (Buffer'Access, Num_Trozo);
         LLU.Send (Server_EP, Buffer'Access);
         Put_Line ("Mando asentimiento del trozo " & Integer'Image (Num_Trozo));
      end loop;


      S_IO.Close (Dst_File);

   end Lectura;









   -- **************************************************************************************************
   -- *********************************** PROGRAMA PRINCIPAL *******************************************
   -- **************************************************************************************************


begin

    LLU.Set_Faults_Vector ( (True, True, True, False, True, False, False, True, False, True) );

   -- CONSTRUYO LA DIRECCIN IP DEL SERVIDOR

   Server_EP := LLU.Build("127.0.0.1", 6123);

   -- ME ATO A UNA DIRECCIN UNICAST PARA EL CLIENTE

   LLU.Bind_Any(Client_EP);

   -- PIDO AL USUARIO SI QUIERE LEER O ESCRIBIR

   Peticion_EL;

   -- PIDO AL USUARIO EL NOMBRE DEL FICHERO

   Peticion_Fichero;

   -- PETICIN DE LAS OPCIONES

   Peticion_Opciones;


   if (Opcion='E') or (Opcion='e') then -- PETICIN DE ESCRITURA
      Escritura;
   else -- PETICIN DEL LECTURA
      Lectura;
   end if;

   LLU.Unbind (Client_EP);
end Cliente;
